import torch
from torch.utils.data import Dataset, DataLoader
import numpy as np

class NumpyDataset(Dataset):
    def __init__(self, data_path, labels_path, transform, three_channels=False):
        """
        Loads image NumPy arrays and labels NumPy arrays and applies transforms to them in a memory-efficient manner.

        Arguments:
        ----------
        data_path: str
            Path to image data of shape [number_of_images, 1, height, width].

        labels_path: str
            Path to label data of shape [number_of_labels, 1].

        transform: Torchvision transforms
            Augmentations to be applied on the data.

        three_channels: bool
            If True the one-channel image is copied over the other channels, creating a three-channel image.
        """

        self.data = np.load(data_path, mmap_mode='r')
        self.labels = np.load(labels_path, mmap_mode='r')

        self.transform = transform
        self.three_channels = three_channels

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data = self.data[idx]
        if self.three_channels:
            data = np.tile(data, (3, 1, 1)) # copies 2d array to the other 3 channels
        
        data = data.astype(np.float32)
        data = self.transform(torch.from_numpy(data))

        label = torch.tensor(self.labels[idx], dtype=torch.long)

        return data, label

def get_dataloader(data_path, labels_path, augmentations, bs=100, three_channels=False, num_workers=2, shuffle=False):
    """
    Creates PyTorch DataLoaders that load image NumPy arrays and labels NumPy arrays and applies transforms to them in a memory-efficient manner.

    Arguments:
    ----------
    data_path: str
        Path to image data of shape [number_of_images, 1, height, width].

    labels_path: str
        Path to label data of shape [number_of_labels, 1].

    transform: Torchvision transforms
        Augmentations to be applied on the data.

    bs: int
        Batch size for loading the data.

    three_channels: bool
        If True the one-channel image is copied over the other channels, creating a three-channel image.

    num_workers: int
        Number of workers to be used (a number too high may slow down loading).

    shuffle: bool
        If True shuffles the data in the DataLoader.
        Some algorithms (Self-Ensemble & AdaMatch) require different DataLoaders that are not shuffled between one another!
    """

    # create dataset
    dataset = NumpyDataset(data_path, labels_path, augmentations, three_channels)
    
    # create dataloader
    dataloader = DataLoader(dataset, shuffle=shuffle, batch_size=bs, num_workers=num_workers)
    
    return dataloader


import torch
from torch import nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
import matplotlib.pyplot as plt
import sklearn.metrics
import seaborn as sns

class Supervised(object):
    def __init__(self, encoder, classifier):
        """
        Arguments:
        ----------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
        """

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.encoder = encoder.to(self.device)
        self.classifier = classifier.to(self.device)

    def train(self, dataloader, dataloader_test, epochs, hyperparams, save_path):
        """
        Trains the model (encoder + classifier).

        Arguments:
        ----------
        dataloader: PyTorch DataLoader
            DataLoader with training data.

        dataloader_test: PyTorch DataLoader
            DataLoader with validation data, used for early stopping.

        epochs: int
            Amount of epochs to train the model for.

        hyperparams: dict
            Dictionary containing hyperparameters for this algorithm. Check `data/hyperparams.py`.

        save_path: str
            Path to store model weights.

        Returns:
        --------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
        """
        
        # configure hyperparameters
        criterion = nn.CrossEntropyLoss()
        lr = hyperparams['learning_rate']
        wd = hyperparams['weight_decay']
        cyclic_scheduler = hyperparams['cyclic_scheduler']

        # concatenate parameters of both self.encoder and self.classifier
        optimizer = optim.Adam(list(self.encoder.parameters()) + list(self.classifier.parameters()), lr=lr, weight_decay=wd)
        if cyclic_scheduler:
            scheduler = optim.lr_scheduler.OneCycleLR(optimizer, lr, epochs=epochs, steps_per_epoch=len(dataloader))

        # early stopping variables
        start_epoch = 0
        best_acc = 0.0
        patience = 15
        bad_epochs = 0

        # metrics
        self.history = {'loss': [], 'accuracy': []}

        # training loop
        for epoch in range(start_epoch, epochs):
            # set network to training mode
            self.encoder.train()
            self.classifier.train()

            running_loss = 0.0

            for data, labels in dataloader:
                data = data.to(self.device)
                labels = labels.to(self.device)

                # zero gradients
                optimizer.zero_grad()

                # classify and get loss
                outputs = self.classifier(self.encoder(data)) # Cs(Ms(xs))
                loss = criterion(outputs, labels)

                # backpropagate and update weights
                loss.backward()
                optimizer.step()

                if cyclic_scheduler:
                    scheduler.step()

                # metrics
                running_loss += loss.item()

            # get losses
            epoch_loss = running_loss / len(dataloader)
            self.history['loss'].append(epoch_loss)

            # evaluate on training data (source domain)
            epoch_accuracy = self.evaluate( dataloader)
            test_epoch_accuracy = self.evaluate(dataloader_test)
            self.history['accuracy'].append(epoch_accuracy)
            
            # save checkpoint
            if test_epoch_accuracy > best_acc:
                torch.save({'encoder_weights': self.encoder.state_dict(),
                            'classifier_weights': self.classifier.state_dict()
                        }, save_path)
                best_acc = test_epoch_accuracy
                bad_epochs = 0
                
            else:
                bad_epochs += 1
            
            print('[Epoch {}/{}] loss: {:.6f}; accuracy: {:.6f}; val accuracy: {:.6f};'.format(epoch+1, epochs, epoch_loss, epoch_accuracy, test_epoch_accuracy))
            
            if bad_epochs >= patience:
                print(f"reached {bad_epochs} bad epochs, stopping training with best val accuracy of {best_acc}!")
                break

        best = torch.load(save_path)
        self.encoder.load_state_dict(best['encoder_weights'])
        self.classifier.load_state_dict(best['classifier_weights'])
        
        return self.encoder, self.classifier

    def evaluate(self, dataloader, return_lists_roc=False):
        """
        Evaluates model on `dataloader`.

        Arguments:
        ----------
        dataloader: PyTorch DataLoader
            DataLoader with test data.

        return_lists_roc: bool
            If True returns also list of labels, a list of outputs and a list of predictions.
            Useful for some metrics.

        Returns:
        --------
        accuracy: float
            Accuracy achieved over `dataloader`.
        """

        # set network to evaluation mode
        self.encoder.eval()
        self.classifier.eval()

        labels_list = []
        outputs_list = []
        preds_list = []

        with torch.no_grad():
            for data, labels in dataloader:
                data = data.to(self.device)
                labels = labels.to(self.device)

                # predict
                outputs = F.softmax(self.classifier(self.encoder(data)), dim=1)

                # numpify
                labels_numpy = labels.detach().cpu().numpy()
                outputs_numpy = outputs.detach().cpu().numpy() # probs (AUROC)
                
                preds = np.argmax(outputs_numpy, axis=1) # accuracy

                # append
                labels_list.append(labels_numpy)
                outputs_list.append(outputs_numpy)
                preds_list.append(preds)

            labels_list = np.concatenate(labels_list)
            outputs_list = np.concatenate(outputs_list)
            preds_list = np.concatenate(preds_list)

        # metrics
        #auc = sklearn.metrics.roc_auc_score(labels_list, outputs_list, multi_class='ovr')
        accuracy = sklearn.metrics.accuracy_score(labels_list, preds_list)

        if return_lists_roc:
            return accuracy, labels_list, outputs_list, preds_list
            
        return accuracy

    def plot_metrics(self):
        """
        Plots the training metrics (only usable after calling .train()).
        """

        # plot metrics from source
        fig, axs = plt.subplots(1, 2, figsize=(12,5), dpi=200)

        epochs = len(self.history['loss'])

        axs[0].plot(range(1, epochs+1), self.history['loss'])
        axs[0].set_xlabel('Epochs')
        axs[0].set_ylabel('Loss')
        axs[0].set_title('Source loss')

        axs[1].plot(range(1, epochs+1), self.history['accuracy'])
        axs[1].set_xlabel('Epochs')
        axs[1].set_ylabel('Accuracy')
        axs[1].set_title('Source accuracy')

        plt.show()

    def plot_cm_roc(self, dataloader, n_classes=3):
        """
        Plots the confusion matrix and ROC curves of the model on `dataloader`.

        Arguments:
        ----------
        dataloader: PyTorch DataLoader
            DataLoader with test data.

        n_classes: int
            Number of classes.
        """

        cmap = sns.diverging_palette(220, 20, as_cmap=True)

        self.encoder.eval()
        self.classifier.eval()

        accuracy, labels_list, outputs_list, preds_list = self.evaluate(dataloader, return_lists_roc=True)

        # plot confusion matrix
        cm = sklearn.metrics.confusion_matrix(labels_list, preds_list)
        group_counts = ['{0:0.0f}'.format(value) for value in cm.flatten()]
        group_percentages = ['({0:.2%})'.format(value) for value in cm.flatten()/np.sum(cm)]
        labels = [f'{v1}\n{v2}' for v1, v2 in zip(group_counts, group_percentages)]
        labels = np.asarray(labels).reshape(3,3)
        #tn, fp, fn, tp = cm.ravel()

        plt.figure(figsize=(4,4), dpi=200)
        categories = ['no', 'spherical', 'vortex'] #  0:no substructure 1:spherical substructure 2:vortex substructure  
        sns.heatmap(cm, annot=labels, cmap=cmap, fmt="", xticklabels=categories, yticklabels=categories)
        plt.title("Confusion matrix")
        plt.ylabel("Actual label")
        plt.xlabel("Predicted label")
        plt.show()

        # plot roc
        ## one hot encode data
        onehot = np.zeros((labels_list.size, labels_list.max()+1))
        onehot[np.arange(labels_list.size),labels_list] = 1
        onehot = onehot.astype('int')

        fpr = dict()
        tpr = dict()
        roc_auc = dict()

        ## get roc curve and auroc for each class
        for i in range(n_classes):
            fpr[i], tpr[i], _ = sklearn.metrics.roc_curve(onehot[:, i], outputs_list[:, i])
            roc_auc[i] = sklearn.metrics.auc(fpr[i], tpr[i])

        ## get macro average auroc
        all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
        mean_tpr = np.zeros_like(all_fpr)
        for i in range(n_classes):
            mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])

        mean_tpr /= n_classes

        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr
        roc_auc["macro"] = sklearn.metrics.auc(fpr["macro"], tpr["macro"])

        plt.figure(figsize=(6,6), dpi=200)

        plt.plot([0, 1], [0, 1], color='black', linestyle='--')

        c = ['blue', 'green', 'orange']
        for i in range(3):
            plt.plot(fpr[i], tpr[i], label=f"AUC class {i} = {roc_auc[i]:.4f}", color=c[i])

        plt.plot(fpr["macro"], tpr["macro"], label=f"macro-average AUC = {roc_auc['macro']:.4f}", color='deeppink', linewidth=2)
            
        plt.title('Receiver Operating Characteristic (ROC)')
        plt.xlabel('False Positives')
        plt.ylabel('True Positives')
        ax = plt.gca()
        ax.set_aspect('equal')
        plt.legend(loc='lower right')
        plt.show()


from .supervised import Supervised
import torch
from torch import nn
import torch.nn.functional as F
import numpy as np
import torch.optim as optim
import matplotlib.pyplot as plt

class Adamatch(Supervised):
    """
    Paper: AdaMatch: A Unified Approach to Semi-Supervised Learning and Domain Adaptation
    Authors: David Berthelot, Rebecca Roelofs, Kihyuk Sohn, Nicholas Carlini, Alex Kurakin
    """

    def __init__(self, encoder, classifier):
        """
        Arguments:
        ----------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
        """

        super().__init__(encoder, classifier)

    def train(self, source_dataloader_weak, source_dataloader_strong,
              target_dataloader_weak, target_dataloader_strong, target_dataloader_test,
              epochs, hyperparams, save_path):
        """
        Trains the model (encoder + classifier).

        Arguments:
        ----------
        source_dataloader_weak: PyTorch DataLoader
            DataLoader with source domain training data with weak augmentations.

        source_dataloader_strong: PyTorch DataLoader
            DataLoader with source domain training data with strong augmentations.

        target_dataloader_weak: PyTorch DataLoader
            DataLoader with target domain training data with weak augmentations.
            THIS DATALOADER'S BATCH SIZE MUST BE 3 * SOURCE_DATALOADER_BATCH_SIZE.

        target_dataloader_strong: PyTorch DataLoader
            DataLoader with target domain training data with strong augmentations.
            THIS DATALOADER'S BATCH SIZE MUST BE 3 * SOURCE_DATALOADER_BATCH_SIZE. 

        target_dataloader_test: PyTorch DataLoader
            DataLoader with target domain validation data, used for early stopping.

        epochs: int
            Amount of epochs to train the model for.

        hyperparams: dict
            Dictionary containing hyperparameters for this algorithm. Check `data/hyperparams.py`.

        save_path: str
            Path to store model weights.

        Returns:
        --------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
        """

        # configure hyperparameters
        lr = hyperparams['learning_rate']
        wd = hyperparams['weight_decay']
        cyclic_scheduler = hyperparams['cyclic_scheduler']
        tau = hyperparams['tau']
        
        iters = max(len(source_dataloader_weak), len(source_dataloader_strong), len(target_dataloader_weak), len(target_dataloader_strong))

        # mu related stuff
        steps_per_epoch = iters
        total_steps = epochs * steps_per_epoch 
        current_step = 0

        # configure optimizer and scheduler
        optimizer = optim.Adam(list(self.encoder.parameters()) + list(self.classifier.parameters()), lr=lr, weight_decay=wd)
        if cyclic_scheduler:
            scheduler = optim.lr_scheduler.OneCycleLR(optimizer, lr, epochs=epochs, steps_per_epoch=iters)

        # early stopping variables
        start_epoch = 0
        best_acc = 0.0
        patience = 15
        bad_epochs = 0

        self.history = {'epoch_loss': [], 'accuracy_source': [], 'accuracy_target': []}

        # training loop
        for epoch in range(start_epoch, epochs):
            running_loss = 0.0

            # set network to training mode
            self.encoder.train()
            self.classifier.train()

            dataset = zip(source_dataloader_weak, source_dataloader_strong, target_dataloader_weak, target_dataloader_strong)

            # this is where the unsupervised learning comes in, as such, we're not interested in labels
            for (data_source_weak, labels_source), (data_source_strong, _), (data_target_weak, _), (data_target_strong, _) in dataset:
                data_source_weak = data_source_weak.to(self.device)
                labels_source = labels_source.to(self.device)

                data_source_strong = data_source_strong.to(self.device)
                data_target_weak = data_target_weak.to(self.device)
                data_target_strong = data_target_strong.to(self.device)

                # concatenate data (in case of low GPU power this could be done after classifying with the model)
                data_combined = torch.cat([data_source_weak, data_source_strong, data_target_weak, data_target_strong], 0)
                source_combined = torch.cat([data_source_weak, data_source_strong], 0)

                # get source data limit (useful for slicing later)
                source_total = source_combined.size(0)

                # zero gradients
                optimizer.zero_grad()

                # forward pass: calls the model once for both source and target and once for source only
                logits_combined = self.classifier(self.encoder(data_combined))
                logits_source_p = logits_combined[:source_total]

                # from https://github.com/yizhe-ang/AdaMatch-PyTorch/blob/main/trainers/adamatch.py
                self._disable_batchnorm_tracking(self.encoder)
                self._disable_batchnorm_tracking(self.classifier)
                logits_source_pp = self.classifier(self.encoder(source_combined))
                
                self._enable_batchnorm_tracking(self.encoder)
                self._enable_batchnorm_tracking(self.classifier)

                # perform random logit interpolation
                lambd = torch.rand_like(logits_source_p).to(self.device)
                final_logits_source = (lambd * logits_source_p) + ((1-lambd) * logits_source_pp)

                # distribution allignment
                ## softmax for logits of weakly augmented source images
                logits_source_weak = final_logits_source[:data_source_weak.size(0)]
                pseudolabels_source = F.softmax(logits_source_weak, 1)

                ## softmax for logits of weakly augmented target images
                logits_target = logits_combined[source_total:]
                logits_target_weak = logits_target[:data_target_weak.size(0)]
                pseudolabels_target = F.softmax(logits_target_weak, 1)

                ## allign target label distribtion to source label distribution
                expectation_ratio = (1e-6 + torch.mean(pseudolabels_source)) / (1e-6 + torch.mean(pseudolabels_target))
                final_pseudolabels = F.normalize((pseudolabels_target * expectation_ratio), p=2, dim=1) # L2 normalization

                # perform relative confidence thresholding
                row_wise_max, _ = torch.max(pseudolabels_source, dim=1)
                final_sum = torch.mean(row_wise_max, 0)
                
                ## define relative confidence threshold
                c_tau = tau * final_sum

                max_values, _ = torch.max(final_pseudolabels, dim=1)
                mask = (max_values >= c_tau).float()

                # compute loss
                source_loss = self._compute_source_loss(logits_source_weak, final_logits_source[data_source_weak.size(0):], labels_source)
                
                final_pseudolabels = torch.max(final_pseudolabels, 1)[1] # argmax
                target_loss = self._compute_target_loss(final_pseudolabels, logits_target[data_target_weak.size(0):], mask)

                ## compute target loss weight (mu)
                pi = torch.tensor(np.pi, dtype=torch.float).to(self.device)
                step = torch.tensor(current_step, dtype=torch.float).to(self.device)
                mu = 0.5 - torch.cos(torch.minimum(pi, (2*pi*step) / total_steps)) / 2

                ## get total loss
                loss = source_loss + (mu * target_loss)
                current_step += 1

                # backpropagate and update weights
                loss.backward()
                optimizer.step()
                if cyclic_scheduler:
                    scheduler.step()

                # metrics
                running_loss += loss.item()

            # get losses
            epoch_loss = running_loss / iters
            self.history['epoch_loss'].append(epoch_loss)

            # self.evaluate on testing data (target domain)
            epoch_accuracy_source = self.evaluate(source_dataloader_weak)
            epoch_accuracy_target = self.evaluate(target_dataloader_weak)
            test_epoch_accuracy = self.evaluate(target_dataloader_test)
            
            self.history['accuracy_source'].append(epoch_accuracy_source)
            self.history['accuracy_target'].append(epoch_accuracy_target)

            # save checkpoint
            if test_epoch_accuracy > best_acc:
                torch.save({'encoder_weights': self.encoder.state_dict(),
                            'classifier_weights': self.classifier.state_dict()
                            }, save_path)
                best_acc = test_epoch_accuracy
                bad_epochs = 0
                
            else:
                bad_epochs += 1
                
            print('[Epoch {}/{}] loss: {:.6f}; accuracy source: {:.6f}; accuracy target: {:.6f}; val accuracy: {:.6f};'.format(epoch+1, epochs, epoch_loss, epoch_accuracy_source, epoch_accuracy_target, test_epoch_accuracy))
            
            if bad_epochs >= patience:
                print(f"reached {bad_epochs} bad epochs, stopping training with best val accuracy of {best_acc}!")
                break

        best = torch.load(save_path)
        self.encoder.load_state_dict(best['encoder_weights'])
        self.classifier.load_state_dict(best['classifier_weights'])
        
        return self.encoder, self.classifier

    def plot_metrics(self):
        """
        Plots the training metrics (only usable after calling .train()).
        """

        # plot metrics for losses n stuff
        fig, axs = plt.subplots(1, 3, figsize=(18,5), dpi=200)

        epochs = len(self.history['epoch_loss'])

        axs[0].plot(range(1, epochs+1), self.history['epoch_loss'])
        axs[0].set_xlabel('Epochs')
        axs[0].set_ylabel('Loss')
        axs[0].set_title('Entropy loss')

        axs[1].plot(range(1, epochs+1), self.history['accuracy_source'])
        axs[1].set_xlabel('Epochs')
        axs[1].set_ylabel('Accuracy')
        axs[1].set_title('Accuracy on weakly augmented source')

        axs[2].plot(range(1, epochs+1), self.history['accuracy_target'])
        axs[2].set_xlabel('Epochs')
        axs[2].set_ylabel('Accuracy')
        axs[2].set_title('Accuracy on weakly augmented target')      
            
        plt.show()

    @staticmethod
    def _disable_batchnorm_tracking(model):
        def fn(module):
            if isinstance(module, nn.modules.batchnorm._BatchNorm):
                module.track_running_stats = False

        model.apply(fn)

    @staticmethod
    def _enable_batchnorm_tracking(model):
        def fn(module):
            if isinstance(module, nn.modules.batchnorm._BatchNorm):
                module.track_running_stats = True

        model.apply(fn)

    @staticmethod
    def _compute_source_loss(logits_weak, logits_strong, labels):
        """
        Receives logits as input (dense layer outputs with no activation function)
        """
        loss_function = nn.CrossEntropyLoss() # default: `reduction="mean"`
        weak_loss = loss_function(logits_weak, labels)
        strong_loss = loss_function(logits_strong, labels)

        #return weak_loss + strong_loss
        return (weak_loss + strong_loss) / 2

    @staticmethod
    def _compute_target_loss(pseudolabels, logits_strong, mask):
        """
        Receives logits as input (dense layer outputs with no activation function).
        `pseudolabels` are treated as ground truth (standard SSL practice).
        """
        loss_function = nn.CrossEntropyLoss(reduction="none")
        pseudolabels = pseudolabels.detach() # remove from backpropagation

        loss = loss_function(logits_strong, pseudolabels)
        
        return (loss * mask).mean()


from torch import nn
from e2cnn import gspaces
import e2cnn.nn as e2nn

class Encoder(nn.Module):
    def __init__(self, features_size=256, sym_group='Dihyderal', N=2):
        """
        Equivariant neural network receives that images and encodes them into an array of size `features_size`.
        NEED TO DOWNGRADE VERSIONS: !pip install torch==1.4 torchvision==0.5

        Arguments:
        ----------
        features_size: int
            Size of encoded features array.

        sym_group: str ['Dihyderal', 'Circular']
            Kind of symmetry group to consider.

        N: int
            Number of discrete rotations (integer greater than 1) or -1 for continuous rotations.
        """

        super(Encoder, self).__init__()

        if sym_group == 'Dihyderal':
            self.r2_act = gspaces.FlipRot2dOnR2(N=N)

        elif sym_group == 'Circular':
            self.r2_act = gspaces.Rot2dOnR2(N=N)

        in_type = e2nn.FieldType(self.r2_act, [self.r2_act.trivial_repr])
        self.input_type = in_type

        out_type = e2nn.FieldType(self.r2_act, 24*[self.r2_act.regular_repr])
        self.block1 = e2nn.SequentialModule(
            e2nn.MaskModule(in_type, 150, margin=1),
            e2nn.R2Conv(in_type, out_type, kernel_size=7, padding=1, bias=False),
            e2nn.InnerBatchNorm(out_type),
            e2nn.ReLU(out_type, inplace=True)
        )

        in_type = self.block1.out_type
        out_type = e2nn.FieldType(self.r2_act, 48*[self.r2_act.regular_repr])
        self.block2 = e2nn.SequentialModule(
            e2nn.R2Conv(in_type, out_type, kernel_size=5, padding=2, bias=False),
            e2nn.InnerBatchNorm(out_type),
            e2nn.ReLU(out_type, inplace=True)
        )
        self.pool1 = e2nn.SequentialModule(
            e2nn.PointwiseAvgPoolAntialiased(out_type, sigma=0.66, stride=2)
        )

        in_type = self.block2.out_type
        out_type = e2nn.FieldType(self.r2_act, 48*[self.r2_act.regular_repr])
        self.block3 = e2nn.SequentialModule(
            e2nn.R2Conv(in_type, out_type, kernel_size=5, padding=2, bias=False),
            e2nn.InnerBatchNorm(out_type),
            e2nn.ReLU(out_type, inplace=True)
        )

        in_type = self.block3.out_type
        out_type = e2nn.FieldType(self.r2_act, 96*[self.r2_act.regular_repr])
        self.block4 = e2nn.SequentialModule(
            e2nn.R2Conv(in_type, out_type, kernel_size=5, padding=2, bias=False),
            e2nn.InnerBatchNorm(out_type),
            e2nn.ReLU(out_type, inplace=True)
        )
        self.pool2 = e2nn.SequentialModule(
            e2nn.PointwiseAvgPoolAntialiased(out_type, sigma=0.66, stride=2)
        )

        in_type = self.block4.out_type
        out_type = e2nn.FieldType(self.r2_act, 96*[self.r2_act.regular_repr])
        self.block5 = e2nn.SequentialModule(
            e2nn.R2Conv(in_type, out_type, kernel_size=5, padding=2, bias=False),
            e2nn.InnerBatchNorm(out_type),
            e2nn.ReLU(out_type, inplace=True)
        )

        in_type = self.block5.out_type
        out_type = e2nn.FieldType(self.r2_act, 64*[self.r2_act.regular_repr])
        self.block6 = e2nn.SequentialModule(
            e2nn.R2Conv(in_type, out_type, kernel_size=5, padding=1, bias=False),
            e2nn.InnerBatchNorm(out_type),
            e2nn.ReLU(out_type, inplace=True)
        )
        self.pool3 = e2nn.PointwiseAvgPoolAntialiased(out_type, sigma=0.66, stride=1, padding=0)

        self.gpool = e2nn.GroupPooling(out_type)
        c = self.gpool.out_type.size

        self.fully_net = nn.Sequential(
            nn.Linear(61504, features_size),
            nn.BatchNorm1d(features_size),
            nn.ELU(inplace=True),
            nn.Linear(features_size, features_size),
        )

    def forward(self, x):
        x = e2nn.GeometricTensor(x, self.input_type)
        x = self.block1(x)
        x = self.block2(x)
        x = self.pool1(x)
        x = self.block3(x)
        x = self.block4(x)
        x = self.pool2(x)
        x = self.block5(x)
        x = self.block6(x)
        x = self.pool3(x)
        x = self.gpool(x)
        x = x.tensor
        x = self.fully_net(x.reshape(x.shape[0], -1))
        return x

class Classifier(nn.Module):
    def __init__(self, features_size=256, num_classes=3):        
        """
        ResNet based neural network that receives an array of size `features_size` and classifies it into `num_classes` classes.

        Arguments:
        ----------
        features_size: int
            Size of encoded features array.

        num_classes: int
            Number of classes to classify the encoded array into.
        """

        super(Classifier, self).__init__()
        self.fc = nn.Linear(features_size, num_classes)

    def forward(self, x):
        x = self.fc(x)
        return x


def supervised_hyperparams(lr=1e-3, wd=1e-5, scheduler=True):
    """
    Return a dictionary of hyperparameters for the Supervised algorithm.
    Default parameters are the best ones as found through a hyperparameter search.

    Arguments:
    ----------
    lr: float
        Learning rate.

    wd: float
        Weight decay for the optimizer.

    scheduler: bool
        Will use a OneCycleLR learning rate scheduler if set to True.

    Returns:
    --------
    hyperparams: dict
        Dictionary containing the hyperparameters. Can be passed to the `hyperparams` argument on Supervised.
    """

    hyperparams = {'learning_rate': lr,
                   'weight_decay': wd,
                   'cyclic_scheduler': scheduler
                   }

    return hyperparams

def adda_hyperparams(lr_target=1e-6, lr_discriminator=1e-5, wd=1e-5, scheduler=False):
    """
    Return a dictionary of hyperparameters for the ADDA algorithm.
    Default parameters are the best ones as found through a hyperparameter search.

    Arguments:
    ----------
    lr_target: float
        Learning rate for the target encoder.

    lr_discriminator: float
        Learning rate for the discriminator.

    wd: float
        Weight decay for the optimizer.

    scheduler: bool
        Will use a OneCycleLR learning rate scheduler if set to True.

    Returns:
    --------
    hyperparams: dict
        Dictionary containing the hyperparameters. Can be passed to the `hyperparams` argument on ADDA.
    """
    
    hyperparams = {'learning_rate_target': lr_target,
                   'learning_rate_discriminator': lr_discriminator,
                   'weight_decay': wd,
                   'cyclic_scheduler': scheduler
                   }

    return hyperparams

def self_ensemble_hyperparams(lr=1e-3, unsupervised_weight=3.0, wd=1e-5, scheduler=False):
    """
    Return a dictionary of hyperparameters for the Self-Ensemble algorithm.
    Default parameters are the best ones as found through a hyperparameter search.

    Arguments:
    ----------
    lr: float
        Learning rate.

    unsupervised_weight: float
        Weight of the unsupervised loss.

    wd: float
        Weight decay for the optimizer.

    scheduler: bool
        Will use a OneCycleLR learning rate scheduler if set to True.

    Returns:
    --------
    hyperparams: dict
        Dictionary containing the hyperparameters. Can be passed to the `hyperparams` argument on ADDA.
    """
    
    hyperparams = {'learning_rate': lr,
                   'unsupervised_weight': unsupervised_weight,
                   'weight_decay': wd,
                   'cyclic_scheduler': scheduler
                   }

    return hyperparams

def cgdm_hyperparams(lr=1e-3, num_k=4, pseudo_interval=2000, wd=1e-5, scheduler=False):
    """
    Return a dictionary of hyperparameters for the CGDM algorithm.
    Default parameters are the best ones as found through a hyperparameter search.

    Arguments:
    ----------
    lr: float
        Learning rate.

    num_k: int
        Amount of steps for the encoder update.

    pseudo_interval: int
        Amount of steps until pseudo labels are updated (smaller interval means more frequent updates).

    wd: float
        Weight decay for the optimizer.

    scheduler: bool
        Will use a OneCycleLR learning rate scheduler if set to True.

    Returns:
    --------
    hyperparams: dict
        Dictionary containing the hyperparameters. Can be passed to the `hyperparams` argument on Self-Ensemble.
    """
    
    hyperparams = {'learning_rate': lr,
                   'num_k': num_k,
                   'pseudo_interval': pseudo_interval,
                   'weight_decay': wd,
                   'cyclic_scheduler': scheduler
                   }

    return hyperparams

def adamatch_hyperparams(lr=1e-3, tau=0.9, wd=1e-5, scheduler=False):
    """
    Return a dictionary of hyperparameters for the AdaMatch algorithm.
    Default parameters are the best ones as found through a hyperparameter search.

    Arguments:
    ----------
    lr: float
        Learning rate.

    tau: float
        Weight of the unsupervised loss.

    wd: float
        Weight decay for the optimizer.

    scheduler: bool
        Will use a OneCycleLR learning rate scheduler if set to True.

    Returns:
    --------
    hyperparams: dict
        Dictionary containing the hyperparameters. Can be passed to the `hyperparams` argument on AdaMatch.
    """
    
    hyperparams = {'learning_rate': lr,
                   'tau': tau,
                   'weight_decay': wd,
                   'cyclic_scheduler': scheduler
                   }

    return hyperparams


from .supervised import Supervised
import torch
from torch import nn
import torch.optim as optim
from torch.autograd import Variable
import matplotlib.pyplot as plt

class Adda(Supervised):
    """
    Paper: Adversarial Discriminative Domain Adaptation
    Authors: Eric Tzeng, Judy Hoffman, Kate Saenko, Trevor Darrell
    """

    def __init__(self, source_encoder, target_encoder, classifier, discriminator):
        """
        Arguments:
        ----------
        source_encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.
            Must have been previously trained in a supervised manner on the source dataset.

        target_encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.
            Must have the same architecture as `source_encoder`.

        classifier: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.

        discriminator: PyTorch neural network
            Neural network that receives an array of size X and classifies it into 2 classes.
            It discriminates between encodings of the source domain and the target domain.
        """

        super().__init__(target_encoder, classifier)

        self.discriminator = discriminator.to(self.device)

        # we consider self.encoder = target_encoder
        self.source_encoder = source_encoder.to(self.device)
        self.encoder.load_state_dict(self.source_encoder.state_dict())

        # disbale grad in already trained networks
        for param in self.source_encoder.parameters():
            param.requires_grad = False

        for param in self.classifier.parameters():
            param.requires_grad = False

    def train(self, source_dataloader, target_dataloader, target_dataloader_test, epochs, hyperparams, save_path):
        """
        Trains the model (encoder + classifier) and discriminator.

        Arguments:
        ----------
        source_dataloader: PyTorch DataLoader
            DataLoader with source domain training data.

        target_dataloader: PyTorch DataLoader
            DataLoader with target domain training data.

        target_dataloader_test: PyTorch DataLoader
            DataLoader with target domain validation data, used for early stopping.

        epochs: int
            Amount of epochs to train the model for.

        hyperparams: dict
            Dictionary containing hyperparameters for this algorithm. Check `data/hyperparams.py`.

        save_path: str
            Path to store model weights.

        Returns:
        --------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
        """
        
        # configure hyperparameters
        criterion = nn.CrossEntropyLoss()
        lr_target = hyperparams['learning_rate_target']
        lr_discriminator = hyperparams['learning_rate_discriminator']
        wd = hyperparams['weight_decay']
        cyclic_scheduler = hyperparams['cyclic_scheduler']
        
        iters = max(len(source_dataloader), len(target_dataloader))
        
        self.source_encoder.eval()
        self.classifier.eval()
        
        # configure optimizers and schedulers
        optimizer_target = optim.Adam(self.encoder.parameters(), lr=lr_target, weight_decay=wd)
        optimizer_discriminator = optim.Adam(self.discriminator.parameters(), lr=lr_discriminator, weight_decay=wd)

        if cyclic_scheduler:
            scheduler_target = optim.lr_scheduler.OneCycleLR(optimizer_target, lr_target, epochs=epochs, steps_per_epoch=iters)
            scheduler_discriminator = optim.lr_scheduler.OneCycleLR(optimizer_discriminator, lr_discriminator, epochs=epochs, steps_per_epoch=iters)

        # early stopping variables
        start_epoch = 0
        best_acc = 0.0
        patience = 15
        bad_epochs = 0

        self.history = {'loss_discriminator': [], 'loss_encoder': [], 'accuracy': []}

        # training loop
        for epoch in range(start_epoch, epochs):
            running_loss_discriminator = 0.0
            running_loss_target = 0.0
            
            # set network to training mode
            self.discriminator.train()
            self.encoder.train()

            # this is where the unsupervised learning comes in, as such, we're not interested in labels
            for (data_source, _), (data_target, _) in zip(source_dataloader, target_dataloader):
                data_source = data_source.to(self.device)
                data_target = data_target.to(self.device)

                # useful when creating labels for self.discriminator
                batch_size_source = data_source.size(0)
                batch_size_target = data_target.size(0)

                # train self.discriminator
                ## zero gradients
                optimizer_discriminator.zero_grad()

                ## encode features and concatenate them
                features_source = self.source_encoder(data_source)
                features_target = self.encoder(data_target)

                ## generate real and fake labels for self.discriminator
                label_source = Variable(torch.zeros(batch_size_source, dtype=torch.long)).to(self.device)
                label_target = Variable(torch.ones(batch_size_target, dtype=torch.long)).to(self.device)
                label_concat = torch.cat([label_source, label_target], dim=0)

                ## classify with self.discriminator
                outputs_source = self.discriminator(features_source)
                outputs_target = self.discriminator(features_target)
                outputs_concat = torch.cat([outputs_source, outputs_target], dim=0)

                ## get loss for self.discriminator
                loss_discriminator = criterion(outputs_concat, label_concat)

                ## backpropagate and update weights
                loss_discriminator.backward()
                optimizer_discriminator.step()

                # train target encoder
                ## zero gradients
                optimizer_discriminator.zero_grad()
                optimizer_target.zero_grad()

                ## encode target features
                features_target = self.encoder(data_target)

                ## classify encoding with self.discriminator
                outputs_target = self.discriminator(features_target)

                ## get loss for target encoder
                label_target = Variable(torch.zeros(batch_size_target, dtype=torch.long)).to(self.device)
                loss_encoder = criterion(outputs_target, label_target)
                
                ## backpropagate and update weights
                loss_encoder.backward()
                optimizer_target.step()

                # metrics
                running_loss_discriminator += loss_discriminator.item()
                running_loss_target += loss_encoder.item()
                
                # scheduler step
                if cyclic_scheduler:
                    scheduler_discriminator.step()
                    scheduler_target.step()

            # get losses
            epoch_loss_discriminator = running_loss_discriminator / iters
            epoch_loss_target = running_loss_target / iters
            self.history['loss_discriminator'].append(epoch_loss_discriminator)
            self.history['loss_encoder'].append(epoch_loss_target)

            # self.evaluate on testing data (target domain)
            epoch_accuracy = self.evaluate(target_dataloader)
            test_epoch_accuracy = self.evaluate(target_dataloader_test)
            self.history['accuracy'].append(epoch_accuracy)

            # save checkpoint
            if test_epoch_accuracy > best_acc:
                torch.save({'encoder_weights': self.encoder.state_dict(),
                            'classifier_weights': self.classifier.state_dict()
                        }, save_path)
                best_acc = test_epoch_accuracy
                bad_epochs = 0
                
            else:
                bad_epochs += 1
                
            print('[Epoch {}/{}] discriminator loss: {:.6f}; target loss: {:.6f}; accuracy target: {:.6f}; val accuracy: {:.6f};'.format(epoch+1, epochs, epoch_loss_discriminator, epoch_loss_target, epoch_accuracy, test_epoch_accuracy))
            
            if bad_epochs >= patience:
                print(f"reached {bad_epochs} bad epochs, stopping training with best val accuracy of {best_acc}!")
                break

        best = torch.load(save_path)
        self.encoder.load_state_dict(best['encoder_weights'])
        self.classifier.load_state_dict(best['classifier_weights'])
        
        return self.encoder, self.classifier

    def plot_metrics(self):
        """
        Plots the training metrics (only usable after calling .train()).
        """

        # plot metrics from target
        fig, axs = plt.subplots(1, 3, figsize=(18,5), dpi=200)

        epochs = len(self.history['loss_discriminator'])

        axs[0].plot(range(1, epochs+1), self.history['loss_discriminator'])
        axs[0].set_xlabel('Epochs')
        axs[0].set_ylabel('Loss')
        axs[0].set_title('Target discriminator loss')

        axs[2].plot(range(1, epochs+1), self.history['loss_encoder'])
        axs[2].set_xlabel('Epochs')
        axs[2].set_ylabel('Loss')
        axs[2].set_title('Target encoder loss')      

        axs[1].plot(range(1, epochs+1), self.history['accuracy'])
        axs[1].set_xlabel('Epochs')
        axs[1].set_ylabel('Accuracy')
        axs[1].set_title('Target accuracy')
            
        plt.show()


from .supervised import Supervised
import torch
from torch import nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
import matplotlib.pyplot as plt
import sklearn.metrics
from scipy.spatial.distance import cdist # cosine distance
from torch.autograd import grad

class Cgdm(Supervised):
    """
    Paper: Cross-Domain Gradient Discrepancy Minimization for Unsupervised Domain Adaptation
    Authors: Zhekai Du, Jingjing Li, Hongzu Su, Lei Zhu, Ke Lu
    """

    def __init__(self, encoder, classifier1, classifier2):
        """
        Here, `classifier1` and `classifier2` are the bi-classifiers.
        They can have different parameters, but must have the same architecture.

        Arguments:
        ----------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier1: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
            If using transfer learning from the source, this should be the source classifier.

        classifier2: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
            If using transfer learning from the source, this should ALSO be the source classifier.
        """

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.encoder = encoder.to(self.device)

        self.classifier1 = classifier1.to(self.device)
        self.classifier2 = classifier2.to(self.device)

    def train(self, source_dataloader, target_dataloader, target_dataloader_test, epochs, hyperparams, save_path):
        """
        Trains the model (encoder + classifier1 + classifier2).

        Arguments:
        ----------
        source_dataloader: PyTorch DataLoader
            DataLoader with source domain training data.

        target_dataloader: PyTorch DataLoader
            DataLoader with target domain training data.

        target_dataloader_test: PyTorch DataLoader
            DataLoader with target domain validation data, used for early stopping.

        epochs: int
            Amount of epochs to train the model for.

        hyperparams: dict
            Dictionary containing hyperparameters for this algorithm. Check `data/hyperparams.py`.

        n_classes: int
            Number of classes.

        save_path: str
            Path to store model weights.

        Returns:
        --------
        encoder: PyTorch neural network
            Neural network that receives images and encodes them into an array of size X.

        classifier1: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.

        classifier2: PyTorch neural network
            Neural network that receives an array of size X and classifies it into N classes.
        """

        # configure hyperparameters
        criterion = nn.CrossEntropyLoss()
        criterion_weighted = self._weighted_crossentropy

        lr = hyperparams['learning_rate']
        wd = hyperparams['weight_decay']
        num_k = hyperparams['num_k']
        cyclic_scheduler = hyperparams['cyclic_scheduler']
        pseudo_interval = hyperparams['pseudo_interval']
        
        iters = max(len(source_dataloader), len(target_dataloader))

        # configure optimizer and scheduler
        optimizer_encoder = optim.Adam(list(self.encoder.parameters()), lr=lr, weight_decay=wd)
        optimizer_classifiers = optim.Adam(list(self.classifier1.parameters()) + list(self.classifier2.parameters()), lr=lr, weight_decay=wd)

        if cyclic_scheduler:
            scheduler_encoder = optim.lr_scheduler.OneCycleLR(optimizer_encoder, lr, epochs=epochs, steps_per_epoch=iters)
            scheduler_classifiers = optim.lr_scheduler.OneCycleLR(optimizer_classifiers, lr, epochs=epochs, steps_per_epoch=iters)

        # early stopping variables
        start_epoch = 0
        best_acc = 0.0
        patience = 15
        bad_epochs = 0

        self.history = {'epoch_loss_entropy': [], 'epoch_loss_classifier1': [], 'epoch_loss_classifier2': [], 'epoch_loss_discrepancy': [], 'accuracy_source': [], 'accuracy_target': []}

        # training loop
        for epoch in range(start_epoch, epochs):
            running_loss_entropy = 0.0
            running_loss_classifier1 = 0.0
            running_loss_classifier2 = 0.0
            running_loss_discrepancy = 0.0

            # this is where the unsupervised learning comes in, as such, we're not interested in labels
            for i, ((data_source, labels_source), (data_target, _)) in enumerate(zip(source_dataloader, target_dataloader)):
                if epoch > start_epoch and i % pseudo_interval == 0:
                    pseudo_labels_target = self._get_pseudo_labels(target_dataloader)

                # set network to training mode
                self.encoder.train()
                self.classifier1.train()
                self.classifier2.train()

                data_source = data_source.to(self.device)
                labels_source = labels_source.to(self.device)

                data_target = data_target.to(self.device)
                if epoch > start_epoch:
                    labels_target = pseudo_labels_target[target_dataloader.batch_size*i : target_dataloader.batch_size*(i+1)]
                    labels_target = labels_target.to(self.device)

                # all steps have similar begginings
                for phase in [1, 2, 3]:
                    for k in range(num_k): # amount of steps to repeat the self.encoder update
                        # zero gradients
                        optimizer_encoder.zero_grad()
                        optimizer_classifiers.zero_grad()
                        
                        # classify the data
                        features_source = self.encoder(data_source)
                        features_target = self.encoder(data_target)

                        outputs1_source = self.classifier1(features_source)
                        outputs1_target = self.classifier1(features_target)
                        outputs2_source = self.classifier2(features_source)
                        outputs2_target = self.classifier2(features_target)

                        # get losses
                        entropy_loss = self._entropy(outputs1_target) + self._entropy(outputs2_target)

                        loss1 = criterion(outputs1_source, labels_source)
                        loss2 = criterion(outputs2_source, labels_source)

                        if phase == 1:
                            # train networks to minimize loss on source
                            if epoch > start_epoch:
                                supervised_loss = criterion_weighted(outputs1_target, labels_target) + criterion_weighted(outputs2_target, labels_target)

                            else:
                                supervised_loss = 0

                            loss = loss1 + loss2 + (0.01 * entropy_loss) + (0.01 * supervised_loss)

                            # backpropagate and update weights
                            loss.backward()
                            optimizer_encoder.step()
                            optimizer_classifiers.step()

                            # exit self.encoder loop (num_k)
                            break

                        elif phase == 2:
                            # train classifiers to maximize divergence between classifier outputs on target (without labels)
                            discrepancy_loss = self._discrepancy(outputs1_target, outputs2_target)
                            loss = loss1 + loss2 - (1.0 * discrepancy_loss) + (0.01 * entropy_loss) 
                            
                            # backpropagate and update weights
                            loss.backward()
                            optimizer_classifiers.step()

                            # exit self.encoder loop (num_k)
                            break

                        elif phase == 3:
                            # train self.encoder to minimize divergence between classifier outputs with gradient similarity
                            discrepancy_loss = self._discrepancy(outputs1_target, outputs2_target)

                            if epoch > start_epoch:
                                source_pack = (outputs1_source, outputs2_source, labels_source)
                                target_pack = (outputs1_target, outputs2_target, labels_target)
                                gradient_discrepancy_loss = self._gradient_discrepancy(source_pack, target_pack)
                            else:
                                gradient_discrepancy_loss = 0

                            loss = (1.0 * discrepancy_loss) + (0.01 * entropy_loss) + (0.01 * gradient_discrepancy_loss)

                            # backpropagate and update weights
                            loss.backward()
                            optimizer_encoder.step()

                # metrics
                running_loss_entropy += entropy_loss.item()
                running_loss_classifier1 += loss1.item()
                running_loss_classifier2 += loss2.item()
                running_loss_discrepancy += discrepancy_loss.item()
                
                # scheduler
                if cyclic_scheduler:
                    scheduler_encoder.step()
                    scheduler_classifiers.step()

            # get losses
            epoch_loss_entropy = running_loss_entropy / iters
            epoch_loss_classifier1 = running_loss_classifier1 / iters
            epoch_loss_classifier2 = running_loss_classifier2 / iters
            epoch_loss_discrepancy = running_loss_discrepancy / iters

            self.history['epoch_loss_entropy'].append(epoch_loss_entropy)
            self.history['epoch_loss_classifier1'].append(epoch_loss_classifier1)
            self.history['epoch_loss_classifier2'].append(epoch_loss_classifier2)
            self.history['epoch_loss_discrepancy'].append(epoch_loss_discrepancy)

            # self.evaluate on training data
            epoch_accuracy_source = self.evaluate(source_dataloader)
            epoch_accuracy_target = self.evaluate(target_dataloader)
            test_epoch_accuracy = self.evaluate(target_dataloader_test)

            self.history['accuracy_source'].append(epoch_accuracy_source)
            self.history['accuracy_target'].append(epoch_accuracy_target)

            # save checkpoint
            if test_epoch_accuracy > best_acc:
                torch.save({'encoder_weights': self.encoder.state_dict(),
                            'classifier1_weights': self.classifier1.state_dict(),
                            'classifier2_weights': self.classifier2.state_dict()
                            }, save_path)
                best_acc = test_epoch_accuracy
                bad_epochs = 0
                
            else:
                bad_epochs += 1

            print('[Epoch {}/{}] entropy loss: {:.6f}; classifier 1 loss: {:.6f}; classifier 2 loss: {:.6f}; discrepancy loss: {:.6f}'.format(epoch+1, epochs, epoch_loss_entropy, epoch_loss_classifier1, epoch_loss_classifier2, epoch_loss_discrepancy))
            print('[Epoch {}/{}] accuracy source: {:.6f}; accuracy target: {:.6f}; val accuracy: {:.6f};'.format(epoch+1, epochs, epoch_accuracy_source, epoch_accuracy_target, test_epoch_accuracy))

            if bad_epochs >= patience:
                print(f"reached {bad_epochs} bad epochs, stopping training with best val accuracy of {best_acc}!")
                break

        best = torch.load(save_path)
        self.encoder.load_state_dict(best['encoder_weights'])
        self.classifier1.load_state_dict(best['classifier1_weights'])
        self.classifier2.load_state_dict(best['classifier2_weights'])

        return self.encoder, self.classifier1, self.classifier2

    def evaluate(self, dataloader, return_lists_roc=False):
        """
        Evaluates model on `dataloader`.

        Arguments:
        ----------
        dataloader: PyTorch DataLoader
            DataLoader with test data.

        return_lists_roc: bool
            If True returns also list of labels, a list of outputs and a list of predictions.
            Useful for some metrics.

        Returns:
        --------
        accuracy: float
            Accuracy achieved over `dataloader`.
        """

        # set network to evaluation mode
        self.encoder.eval()
        self.classifier1.eval()
        self.classifier2.eval()

        labels_list = []
        outputs_list = []
        preds_list = []

        with torch.no_grad():
            for data, labels in dataloader:
                data = data.to(self.device)
                labels = labels.to(self.device)

                # predict
                features = self.encoder(data)
                outputs1 = self.classifier1(features)
                outputs2 = self.classifier2(features)
                
                outputs = F.softmax(outputs1 + outputs2, dim=1)

                # numpify
                labels_numpy = labels.detach().cpu().numpy()
                outputs_numpy = outputs.detach().cpu().numpy() # probs (AUROC)
                
                preds = np.argmax(outputs_numpy, axis=1) # accuracy

                # append
                labels_list.append(labels_numpy)
                outputs_list.append(outputs_numpy)
                preds_list.append(preds)

            labels_list = np.concatenate(labels_list)
            outputs_list = np.concatenate(outputs_list)
            preds_list = np.concatenate(preds_list)

        # metrics
        accuracy = sklearn.metrics.accuracy_score(labels_list, preds_list)

        if return_lists_roc:
            return accuracy, labels_list, outputs_list, preds_list
            
        return accuracy

    def plot_metrics(self):
        """
        Plots the training metrics (only usable after calling .train()).
        """

        # plot metrics for losses n stuff
        fig, axs = plt.subplots(1, 3, figsize=(18,5), dpi=200)

        epochs = len(self.history['epoch_loss_entropy'])

        axs[0].plot(range(1, epochs+1), self.history['epoch_loss_entropy'])
        axs[0].set_xlabel('Epochs')
        axs[0].set_ylabel('Loss')
        axs[0].set_title('Entropy loss')

        axs[1].plot(range(1, epochs+1), self.history['epoch_loss_classifier1'], label='classifier 1')
        axs[1].plot(range(1, epochs+1), self.history['epoch_loss_classifier2'], label='classifier 2')
        axs[1].set_xlabel('Epochs')
        axs[1].set_ylabel('Loss')
        axs[1].set_title('Classifier loss')
        axs[1].legend()

        axs[2].plot(range(1, epochs+1), self.history['epoch_loss_discrepancy'])
        axs[2].set_xlabel('Epochs')
        axs[2].set_ylabel('Counts')
        axs[2].set_title('Discrepancy loss')      
            
        plt.show()

        fig, axs = plt.subplots(1, 2, figsize=(12,5), dpi=200)

        axs[0].plot(range(1, epochs+1), self.history['accuracy_source'])
        axs[0].set_xlabel('Epochs')
        axs[0].set_ylabel('Accuracy')
        axs[0].set_title('Accuracy on source')

        axs[1].plot(range(1, epochs+1), self.history['accuracy_target'])
        axs[1].set_xlabel('Epochs')
        axs[1].set_ylabel('Accuracy')
        axs[1].set_title('Accuracy on target')

        plt.show()

    @staticmethod
    def _weighted_crossentropy(input, target):
        input_softmax = F.softmax(input, dim=1)
        entropy = -input_softmax * torch.log(input_softmax + 1e-5) # standard info entropy with "anti-zero" term
        entropy = torch.sum(entropy, dim=1)

        weight = 1.0 + torch.exp(-entropy)
        weight = weight / torch.sum(weight).detach().item()

        return torch.mean(weight * F.cross_entropy(input, target, reduction='none'))

    @staticmethod
    def _entropy(input, epsilon=1e-5):
        # apply softmax
        input = F.softmax(input, dim=1)
        
        # entropy_condition
        entropy_condition = -input * torch.log(input + epsilon)
        entropy_condition = torch.sum(entropy_condition, dim=1).mean()
        
        # entropy_div
        input = torch.mean(input, 0) + epsilon
        entropy_div = input * torch.log(input)
        entropy_div = torch.sum(entropy_div)

        return entropy_condition + entropy_div

    @staticmethod
    def _discrepancy(input1, input2):
        return torch.mean(torch.abs(F.softmax(input1, dim=1) - F.softmax(input2, dim=1)))

    def _gradient_discrepancy(self, source_pack, target_pack):
        outputs1_source, outputs2_source, labels_source = source_pack
        outputs1_target, outputs2_target, labels_target = target_pack

        criterion = nn.CrossEntropyLoss()
        criterion_weighted = self._weighted_crossentropy

        gradient_loss = 0

        # get losses
        loss1_source = criterion(outputs1_source, labels_source)
        loss2_source = criterion(outputs2_source, labels_source)
        losses_source = [loss1_source, loss2_source]

        loss1_target = criterion_weighted(outputs1_target, labels_target)
        loss2_target = criterion_weighted(outputs2_target, labels_target)
        losses_target = [loss1_target, loss2_target]

        # get gradient loss from each classifier
        for classifier, loss_source, loss_target in zip([self.classifier1, self.classifier2], losses_source, losses_target):
            grad_cosine_similarity = []
            
            for name, params in classifier.named_parameters():
                real_grad = grad([loss_source], [params], create_graph=True, only_inputs=True, allow_unused=False)[0]
                fake_grad = grad([loss_target], [params], create_graph=True, only_inputs=True, allow_unused=False)[0]

                if len(params.shape) > 1:
                    cosine_similarity = F.cosine_similarity(fake_grad, real_grad, dim=1).mean()
                else:
                    cosine_similarity = F.cosine_similarity(fake_grad, real_grad, dim=0)

                grad_cosine_similarity.append(cosine_similarity)

            # concatenate cosine similarities
            grad_cosine_similarity = torch.stack(grad_cosine_similarity)

            # get loss for this classifier
            gradient_loss += (1.0 - grad_cosine_similarity).mean()

        return gradient_loss/2.0 # mean of both gradient_loss(es)

    def _get_pseudo_labels(self, target_dataloader):
        self.encoder.eval()
        self.classifier1.eval()
        self.classifier2.eval()

        start_test = True

        with torch.no_grad():
            for data, labels in target_dataloader:
                data = data.to(self.device)
                labels = labels.to(self.device)

                # generate features
                features = self.encoder(data)

                outputs1 = self.classifier1(features)
                outputs2 = self.classifier2(features)
                outputs = outputs1 + outputs2

                if start_test:
                    all_features = features.float().cpu()
                    all_outputs = outputs.float().cpu()
                    all_labels = labels.float().cpu()
                    start_test = False
                
                else:
                    all_features = torch.cat((all_features, features.float().cpu()), dim=0)
                    all_outputs = torch.cat((all_outputs, outputs.float().cpu()), dim=0)
                    all_labels = torch.cat((all_labels, labels.float().cpu()), dim=0)

        all_outputs = F.softmax(all_outputs, dim=1)
        _, preds = torch.max(all_outputs, dim=1)
        accuracy = torch.sum(torch.squeeze(preds).float() == all_labels).item() / float(all_labels.size()[0])

        all_features = torch.cat((all_features, torch.ones(all_features.size(0), 1)), dim=1)
        all_features = (all_features.t() / torch.norm(all_features, p=2, dim=1)).t()
        all_features = all_features.float().cpu().numpy()

        # perform k-means clustering
        k = all_outputs.size(1)

        for i in range(2):
            if i == 0:
                aff = all_outputs.float().cpu().numpy()
            else:
                aff = np.eye(k)[preds_label]
            
            initial_centroid = aff.transpose().dot(all_features)
            initial_centroid = initial_centroid / (1e-8 + aff.sum(axis=0)[:,None])
            distance = cdist(all_features, initial_centroid, 'cosine')

            preds_label = distance.argmin(axis=1)
            accuracy_kmeans = np.sum(preds_label == all_labels.float().cpu().numpy()) / len(all_features)

        print('only source accuracy = {:.2f}% -> after clustering = {:.2f}%'.format(accuracy*100, accuracy_kmeans*100))
        return torch.tensor(preds_label, dtype=torch.long)